"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MockManager = void 0;
var defaultMockOptions = {};
var MockManager = /** @class */ (function () {
    function MockManager(provider) {
        this.provider = provider;
    }
    /**
     * Emits an event
     * @param eventName Name of the event
     * @param payload Payload of the event
     * @example ```ts
     * mockManager.emit("chainChanged", "0x1");
     * ```
     */
    MockManager.prototype.emit = function (eventName, payload) {
        this.provider.emit(eventName, payload);
    };
    /**
     * Mock a JSON-RPC request
     * @param method JSON-RPC method name
     * @param data Data to be resolved, or function to be called, or error to be thrown in case of throw
     * @param mockOptions Options of the mock
     * @example ```ts
     * // Mock one "eth_accounts" request
     * mockManager.mockRequest("eth_accounts", ["0x..."]);
     * // Persistently mock "eth_chainId" request
     * mockManager.mockRequest("eth_chainId", "0x1", { persistent: true });
     * // Mock with a dynamical value based on params
     * // "personal_sign" in this case, `bobsWallet` is externally defined here
     * mockManager.mockRequest("personal_sign", async (params: unknown[]) => {
     *   const statement = (params as string[])[0];
     *   return await bobsWallet.signMessage(statement);
     * });
     * ```
     */
    MockManager.prototype.mockRequest = function (method, data, mockOptions) {
        if (mockOptions === void 0) { mockOptions = defaultMockOptions; }
        var condition = mockOptions.condition, persistent = mockOptions.persistent;
        var isConditionalMock = Boolean(condition);
        var isPersistent = Boolean(persistent);
        var mockRequest = __assign({ method: method, data: data }, mockOptions);
        if (isConditionalMock) {
            var currentConditionalPersistentMock_1 = this.findConditionalPersistentMock(method, condition);
            if (currentConditionalPersistentMock_1) {
                if (!isPersistent) {
                    console.warn("There is already a persistent registered mock for ".concat(method, " with this condition, this additional mocking will not be considered."));
                    return this;
                }
                this.provider.requestMocks[method] = this.provider.requestMocks[method].filter(function (mock) { return mock !== currentConditionalPersistentMock_1; });
            }
            return this.registerMock(method, mockRequest);
        }
        var currentUnconditionalPersistentMock = this.findUnconditionalPersistentMock(method);
        if (currentUnconditionalPersistentMock) {
            if (!isPersistent) {
                console.warn("There is already a persistent registered mock for ".concat(method, ", this additional mocking will not be considered."));
                return this;
            }
            this.provider.requestMocks[method] = this.provider.requestMocks[method].filter(function (mock) { return mock !== currentUnconditionalPersistentMock; });
        }
        return this.registerMock(method, mockRequest);
    };
    MockManager.prototype.registerMock = function (method, mockRequest) {
        var mocks = this.provider.requestMocks[method];
        if (!mocks) {
            this.provider.requestMocks[method] = [mockRequest];
            return this;
        }
        mocks.push(mockRequest);
        return this;
    };
    MockManager.prototype.findUnconditionalPersistentMock = function (method) {
        var mocks = this.provider.requestMocks[method];
        if (!mocks)
            return false;
        return mocks.find(function (mock) { return mock.persistent && !Boolean(mock.condition); });
    };
    MockManager.prototype.findConditionalPersistentMock = function (method, condition) {
        var mocks = this.provider.requestMocks[method];
        if (!mocks)
            return false;
        return mocks.find(function (mock) { return mock.persistent && mock.condition === condition; });
    };
    /**
     * Clear all mocks for the provider
     */
    MockManager.prototype.clearAllMocks = function () {
        this.provider.requestMocks = {};
    };
    return MockManager;
}());
exports.MockManager = MockManager;
