"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ContractUtils = void 0;
var ethers_1 = require("ethers");
var abi_1 = require("@ethersproject/abi");
var bignumber_1 = require("@ethersproject/bignumber");
var ContractUtils = /** @class */ (function () {
    function ContractUtils(mockManager, abi, address) {
        this.mockManager = mockManager;
        this.contractInterface = new abi_1.Interface(abi);
        this.address = address;
        this.conditionCache = {};
    }
    /**
     * Mock a call of the contract
     * @param functionName Name of the function
     * @param values Array of values to be returned - array is used as ordering of the returned data
     * @param callOptions.callValues Optional array of values passed to the call - useful when dealing with multiple similar calls with different arguments
     * @param callOptions.contractAddress Optional address of the contract, fallbacks to the contract address of the utils if specified
     * @param mockOptions.persistent If true, the mock will persist
     * @param mockOptions.shouldThrow If true, the mocked request will throw, the thrown error is the data field
     * @param mockOptions.timeout Timeout of the request, in milliseconds
     * @param mockOptions.condition Specific condition function for the mock, a conditional mock has priority over standard mocks. Signature is `(params: unknown[]) => boolean`
     * @param mockOptions.triggerCallback Callback that is triggered when the mock has been consumed. Signature is `(data?: unknown, params?: unknown[]) => void`
     * @example ```ts
     * // Mock next call to `value` function with return value `12`
     * contractUtils.mockCall("value", ["12"]);
     * ```
     */
    ContractUtils.prototype.mockCall = function (functionName, values, callOptions, mockOptions) {
        var _this = this;
        if (callOptions === void 0) { callOptions = {}; }
        if (mockOptions === void 0) { mockOptions = {}; }
        var contractAddress = callOptions.contractAddress, callValues = callOptions.callValues;
        var toAddress = contractAddress || this.address;
        var conditionKey = JSON.stringify({
            functionName: functionName,
            to: toAddress,
            callValues: callValues,
        });
        var condition;
        var conditionFromCache = this.conditionCache[conditionKey];
        if (Boolean(conditionFromCache)) {
            condition = conditionFromCache;
        }
        else {
            condition = function (params) {
                var ethCallParams = params;
                var _a = ethCallParams[0], to = _a.to, data = _a.data;
                var isTargetedContract = toAddress
                    ? toAddress.toLowerCase() === to.toLowerCase()
                    : true;
                var isTargetedFunction = callValues
                    ? data ===
                        _this.contractInterface.encodeFunctionData(functionName, callValues)
                    : data.startsWith(_this.contractInterface.getSighash(functionName));
                return isTargetedContract && isTargetedFunction;
            };
            this.conditionCache[conditionKey] = condition;
        }
        var callResult = this.contractInterface.encodeFunctionResult(functionName, values);
        this.mockManager.mockRequest("eth_call", callResult, __assign({ condition: condition }, mockOptions));
        return this;
    };
    /**
     * Mock a transaction of the contract
     * @param functionName Name of the function
     * @param txOptions.from Optional address of the sender, fallbacks to the current mocked account
     * @param txOptions.to Optional address of the recipient, fallbacks to the contract address of the utils if specified
     * @param txOptions.txValues Optional array of values passed to the transaction - useful when dealing with multiple similar calls with different arguments
     * @param mockOptions.persistent If true, the mock will persist
     * @param mockOptions.shouldThrow If true, the mocked request will throw, the thrown error is the data field
     * @param mockOptions.timeout Timeout of the request, in milliseconds
     * @param mockOptions.condition Specific condition function for the mock, a conditional mock has priority over standard mocks. Signature is `(params: unknown[]) => boolean`
     * @param mockOptions.triggerCallback Callback that is triggered when the mock has been consumed. Signature is `(data?: unknown, params?: unknown[]) => void`
     * @example ```ts
     * // Mock next transaction using `setValue` method
     * contractUtils.mockTransaction("setValue");
     * ```
     */
    ContractUtils.prototype.mockTransaction = function (functionName, txOptions, mockOptions) {
        var _this = this;
        if (txOptions === void 0) { txOptions = {}; }
        if (mockOptions === void 0) { mockOptions = {}; }
        var chainIdMock = this.mockManager.findUnconditionalPersistentMock("eth_chainId");
        if (!chainIdMock) {
            throw new Error("Unable to properly mock transaction because chain ID mock has not been properly set up.");
        }
        var accountMock = this.mockManager.findUnconditionalPersistentMock("eth_accounts");
        if (!accountMock) {
            throw new Error("Unable to properly mock transaction because account mock has not been properly set up.");
        }
        var txValues = txOptions.txValues, from = txOptions.from, to = txOptions.to;
        var blockNumberMock = this.mockManager.findUnconditionalPersistentMock("eth_blockNumber");
        if (!blockNumberMock) {
            throw new Error("Unable to properly mock transaction because block number mock has not been properly set up.");
        }
        var blockNumber = blockNumberMock.data;
        var mockedAccount = accountMock.data[0];
        var fromAddress = from || mockedAccount;
        var toAddress = to || this.address;
        var conditionKey = JSON.stringify({
            functionName: functionName,
            to: toAddress,
            from: fromAddress,
            txValues: txValues,
        });
        var condition;
        var conditionFromCache = this.conditionCache[conditionKey];
        if (Boolean(conditionFromCache)) {
            condition = conditionFromCache;
        }
        else {
            condition = function (params) {
                var estimateGasOrSendTransactionParams = params;
                var _a = estimateGasOrSendTransactionParams[0], from = _a.from, to = _a.to, data = _a.data;
                var isTargetedFunction = txValues
                    ? data ===
                        _this.contractInterface.encodeFunctionData(functionName, txValues)
                    : data.startsWith(_this.contractInterface.getSighash(functionName));
                var isTargetedContract = toAddress
                    ? toAddress.toLowerCase() === to.toLowerCase()
                    : true;
                return (from.toLowerCase() === fromAddress.toLowerCase() &&
                    isTargetedContract &&
                    isTargetedFunction);
            };
            this.conditionCache[conditionKey] = condition;
        }
        var gasEstimation = "0x0000000000000000000000000000000000000000000000000000000000010000";
        this.mockManager.mockRequest("eth_estimateGas", gasEstimation, {
            condition: condition,
        });
        var txHash = "0xdafc0b053f0728212d1a7bf7f12b883107db7a2ef949a28c2483cabaf187255c";
        this.mockManager.mockRequest("eth_sendTransaction", txHash, __assign({ condition: condition }, mockOptions));
        var txData = txValues
            ? this.contractInterface.encodeFunctionData(functionName, txValues)
            : "0x55241077000000000000000000000000000000000000000000000000000000000000007b";
        var tx = {
            to: toAddress,
            from: fromAddress,
            input: txData,
            chainId: "0x1",
            hash: txHash,
            nonce: "0x1",
            gasLimit: bignumber_1.BigNumber.from(1).toHexString(),
            value: bignumber_1.BigNumber.from(0).toHexString(),
            r: "0x",
            s: "0x",
            v: "0x",
            type: "0x1",
            transactionIndex: "0x1",
        };
        var getTxCondition = function (params) {
            var getTxByHashParams = params;
            var hash = getTxByHashParams[0];
            return hash === txHash;
        };
        this.mockManager.mockRequest("eth_getTransactionByHash", tx, {
            condition: getTxCondition,
        });
        var blockHash = "0xdafc0b053f0728212d1a7bf7f12b883107db7a2ef949a28c2483cabaf187255c";
        var defaultContractAddress = "0xf61B443A155b07D2b2cAeA2d99715dC84E839EEf";
        var txReceipt = {
            to: toAddress || defaultContractAddress,
            from: fromAddress,
            contractAddress: toAddress || defaultContractAddress,
            transactionIndex: "0x1",
            gasUsed: bignumber_1.BigNumber.from(1).toHexString(),
            logsBloom: "",
            blockHash: blockHash,
            transactionHash: txHash,
            logs: [],
            blockNumber: blockNumber,
            confirmations: "0x1",
            cumulativeGasUsed: bignumber_1.BigNumber.from(1).toHexString(),
            effectiveGasPrice: bignumber_1.BigNumber.from(1).toHexString(),
            // byzantium: true,
            type: "0x1",
        };
        this.mockManager.mockRequest("eth_getTransactionReceipt", txReceipt, {
            condition: getTxCondition,
        });
        return this;
    };
    /**
     * Mock the next past logs request with an array of a single event type
     * @param eventName Name of the event
     * @param allValues Array of array of values of the events
     * @example ```ts
     * // Mock for two events `event ValueUpdated(uint value)` with values `0` and `12`
     * contractTestingUtils.mockGetLogs("ValueUpdated", [["0"], ["12"]]);
     * ```
     */
    ContractUtils.prototype.mockGetLogs = function (eventName, allValues) {
        var _this = this;
        var blockNumberMock = this.mockManager.findUnconditionalPersistentMock("eth_blockNumber");
        if (!blockNumberMock) {
            throw new Error("Unable to properly mock transaction because block number mock has not been properly set up.");
        }
        var blockNumber = blockNumberMock.data;
        var eventTopic = this.contractInterface.getEventTopic(this.contractInterface.getEvent(eventName));
        var condition = function (params) {
            var _a;
            var filter = params[0];
            return ((_a = filter === null || filter === void 0 ? void 0 : filter.topics) === null || _a === void 0 ? void 0 : _a[0]) === eventTopic;
        };
        this.mockManager.mockRequest("eth_getLogs", allValues.map(function (values) {
            return _this.generateMockLog(eventName, values, {
                blockNumber: blockNumber,
            });
        }), { condition: condition });
        return this;
    };
    /**
     * Mock an emission of a log, mocked block number is automatically increased
     * @param eventName Name of the event
     * @param values array of values of the event
     * @param subscriptionId Needed if using web3.js
     * @param logOverrides Optional overrides for the log
     * @example ```ts
     * // Mock emission of a log for `event ValueUpdated(uint value)` with value `12`
     * ...
     * // With ethers
     * contractTestingUtils.mockEmitLog("ValueUpdated", ["12"]);
     * // With web3.js
     * testingUtils.mockSubscribe("0x123");
     * ...
     * contractTestingUtils.mockEmitLog("ValueUpdated", ["12"], "0x123");
     * ```
     */
    ContractUtils.prototype.mockEmitLog = function (eventName, values, subscriptionId, logOverrides) {
        var blockNumberMock = this.mockManager.findUnconditionalPersistentMock("eth_blockNumber");
        if (!blockNumberMock) {
            throw new Error("Unable to properly mock transaction because block number mock has not been properly set up.");
        }
        var blockNumber = blockNumberMock.data;
        var incrementedBlockNumber = Number(blockNumber) + 1;
        var hexValue = ethers_1.ethers.utils.hexValue(incrementedBlockNumber);
        var zeroPadIncrementedBlockNumber = ethers_1.ethers.utils.hexZeroPad(hexValue, 32);
        this.mockManager.mockRequest("eth_blockNumber", zeroPadIncrementedBlockNumber, { persistent: true });
        var log = this.generateMockLog(eventName, values, __assign({ blockNumber: "0x".concat(incrementedBlockNumber.toString(16)) }, logOverrides));
        if (subscriptionId) {
            this.mockManager.emit("message", {
                type: "eth_subscription",
                data: {
                    subscription: subscriptionId,
                    result: log,
                },
            });
        }
        else {
            var eventTopic_1 = this.contractInterface.getEventTopic(this.contractInterface.getEvent(eventName));
            var condition = function (params) {
                var _a;
                var filter = params[0];
                return ((_a = filter === null || filter === void 0 ? void 0 : filter.topics) === null || _a === void 0 ? void 0 : _a[0]) === eventTopic_1;
            };
            this.mockManager.mockRequest("eth_getLogs", [log], { condition: condition });
            var address = this.address || (logOverrides === null || logOverrides === void 0 ? void 0 : logOverrides.address);
            var filters = {
                address: address,
                topics: this.contractInterface.encodeFilterTopics(this.contractInterface.getEvent(eventName), []),
            };
            this.mockManager.emit(JSON.stringify(filters), log);
        }
        return this;
    };
    /**
     * Generate a log from an event name and the values of the event
     * @param eventName Name of the event
     * @param values Array of values of the event
     * @param logOverrides Optional overrides for the log
     * @example ```ts
     * // Generate a log for `event ValueUpdated(uint value)` with value `12`
     * const log = contractTestingUtils.generateMockLog("ValueUpdated", ["12"]);
     * ```
     * @returns The log for the event
     */
    ContractUtils.prototype.generateMockLog = function (eventName, values, logOverrides) {
        var defaultContractAddress = "0xf61B443A155b07D2b2cAeA2d99715dC84E839EEf";
        var address = this.address || defaultContractAddress;
        var _a = this.contractInterface.encodeEventLog(this.contractInterface.getEvent(eventName), values), data = _a.data, topics = _a.topics;
        return __assign({ blockNumber: "0x1", blockHash: "0xdafc0b053f0728212d1a7bf7f12b883107db7a2ef949a28c2483cabaf187255c", transactionIndex: "0x0", removed: false, address: address, data: data, topics: topics, transactionHash: "0xdafc0b053f0728212d1a7bf7f12b883107db7a2ef949a28c2483cabaf187255c", logIndex: "0x0" }, logOverrides);
    };
    return ContractUtils;
}());
exports.ContractUtils = ContractUtils;
