"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EnsUtils = void 0;
var ethers_1 = require("ethers");
var __1 = require("..");
var constants_1 = require("./constants");
var EnsUtils = /** @class */ (function () {
    function EnsUtils(mockManager) {
        this.ensRegistryWithFallbackUtils = new __1.ContractUtils(mockManager, constants_1.ENS_REGISTRY_WITH_FALLBACK_ABI, constants_1.ENS_REGISTRY_WITH_FALLBACK_ADDRESS);
        this.publicResolverUtils = new __1.ContractUtils(mockManager, constants_1.PUBLIC_RESOLVER_ABI, constants_1.PUBLIC_RESOLVER_ADDRESS);
    }
    /**
     * Mock the resolver address of a name
     * @param name ENS name
     * @param resolverAddress Resolver address, fallback to the main net public resolver
     * @example ```ts
     * ens.mockResolver("blabla.eth");
     * ```
     */
    EnsUtils.prototype.mockResolver = function (name, resolverAddress) {
        if (resolverAddress === void 0) { resolverAddress = constants_1.PUBLIC_RESOLVER_ADDRESS; }
        var callValues = name
            ? {
                callValues: [ethers_1.ethers.utils.namehash(name)],
            }
            : undefined;
        this.ensRegistryWithFallbackUtils.mockCall("resolver", [resolverAddress], callValues, { persistent: true });
        return this;
    };
    /**
     * Mock all addresses to resolve to empty names and names to zero address
     * @example ```ts
     * ens.mockAllToEmpty();
     * ```
     */
    EnsUtils.prototype.mockAllToEmpty = function () {
        this.mockResolver();
        this.publicResolverUtils.mockCall("supportsInterface", [false], undefined, {
            persistent: true,
        });
        this.publicResolverUtils.mockCall("name", [""], undefined, {
            persistent: true,
        });
        this.publicResolverUtils.mockCall("addr(bytes32)", [ethers_1.ethers.constants.AddressZero], undefined, { persistent: true });
    };
    /**
     * Mock empty reverse resolution of a batch of addresses
     * @param addresses Array of address
     * @example ```ts
     * ens.mockEmptyReverse(["0x123...", "0x456..."]);
     * ```
     */
    EnsUtils.prototype.mockEmptyReverse = function (addresses) {
        var _this = this;
        this.publicResolverUtils.mockCall("supportsInterface", [false], undefined, {
            persistent: true,
        });
        addresses.forEach(function (address) {
            var reverseName = address.substring(2).toLowerCase() + ".addr.reverse";
            _this.mockResolver(reverseName, ethers_1.ethers.constants.AddressZero);
            var callValues = {
                callValues: [ethers_1.ethers.utils.namehash(reverseName)],
            };
            _this.publicResolverUtils.mockCall("name", [""], callValues, {
                persistent: true,
            });
        });
    };
    /**
     * Mock the resolution of a name to an address
     * @param name ENS name
     * @param address Address
     * @example ```ts
     * ens.mockResolve("blabla.eth", "0x123...");
     * ```
     */
    EnsUtils.prototype.mockResolve = function (name, address) {
        this.mockResolver(name);
        var callValues = {
            callValues: [ethers_1.ethers.utils.namehash(name)],
        };
        this.publicResolverUtils.mockCall("supportsInterface", [false], undefined, {
            persistent: true,
        });
        this.publicResolverUtils.mockCall("addr(bytes32)", [address], callValues, {
            persistent: true,
        });
    };
    /**
     * Mock the reverse resolution of an address to a name
     * @param address Address
     * @param name ENS name
     * @example ```ts
     * ens.mockReverseResolve("0x123...", "blabla.eth");
     * ```
     */
    EnsUtils.prototype.mockReverseResolve = function (address, name) {
        var reverseName = address.substring(2).toLowerCase() + ".addr.reverse";
        this.mockResolver(reverseName);
        this.publicResolverUtils.mockCall("supportsInterface", [false], undefined, {
            persistent: true,
        });
        var callValues = {
            callValues: [ethers_1.ethers.utils.namehash(reverseName)],
        };
        this.publicResolverUtils.mockCall("name", [name], callValues, {
            persistent: true,
        });
    };
    /**
     * Mock a name to resolve to an address and the address to reverse resolve to the name
     * @param address Address
     * @param name ENS name
     * @example ```ts
     * ens.mockName("0x123...", "blabla.eth");
     * ```
     */
    EnsUtils.prototype.mockEnsName = function (address, name) {
        this.mockResolve(name, address);
        this.mockReverseResolve(address, name);
    };
    return EnsUtils;
}());
exports.EnsUtils = EnsUtils;
